@@ -70,10 +70,14 @@ class ElementType(Enum):
     ARTIST_FOCUS = 51, "rtstrl"
     ALBUMS_IN_MIX_OR_PLAYLIST = 52, "lbmnmxplst"
     ARTISTS_IN_MIX_OR_PLAYLIST = 53, "rtstnmxplst"
-    ALBUM_LISTEN_QUEUE_ACTION = 54, "lbmlqctn"
-    ALBUM_LISTEN_QUEUE = 55, "lbllstnq"
+    BOOKMARK_ALBUM_ACTION = 54, "lbmlqctn"
+    BOOKMARK_ALBUMS = 55, "lbllstnq"
     ALBUM_TRACKS = 56, "lbmtrks"
     TRACK_FAVORITE_ACTION = 57, "trkfvtct"
+    BOOKMARK_ARTISTS = 58, "rtstlstnq"
+    BOOKMARK_ARTIST_ACTION = 59, "rtstlqctn"
+    BOOKMARK_TRACKS = 60, "snglstnq"
+    BOOKMARK_TRACK_ACTION = 61, "snglstnqctn"
     TRACK = 101, "trk"
     TRACK_CONTAINER = 102, "trkc"
 
