@@ -17,6 +17,9 @@ import cmdtalkplugin
 import upmplgutils
 import constants
 import config
+import requests
+import os
+import upnp_util
 
 from tidalapi import Quality as TidalQuality
 from tidalapi.session import Session as TidalSession
@@ -27,11 +30,15 @@ from tidalapi.playlist import UserPlaylist as TidalUserPlaylist
 from tidalapi.mix import Mix as TidalMix
 from tidalapi.media import MediaMetadataTags as TidalMediaMetadataTags
 from tidalapi.media import Track as TidalTrack
+from tidalapi.media import AudioMode as TidalAudioMode
+
 from typing import Callable
 
 from element_type import ElementType
 from element_type import get_element_type_by_name
 
+from album_sort_criteria import AlbumSortCriteria
+
 # Func name to method mapper
 dispatcher = cmdtalkplugin.Dispatch()
 # Pipe message handler
@@ -50,6 +57,44 @@ default_image_sz_by_type[TidalMix.__name__] = [1500, 640, 320]
 default_image_sz_by_type[TidalUserPlaylist.__name__] = default_image_sz_by_type[TidalPlaylist.__name__]
 
 
+class FavoriteAlbumsMode:
+
+    __element_type: ElementType
+    __display_name: str
+    __sort_criteria_builder: Callable[[bool], list[AlbumSortCriteria]]
+    __descending: bool
+
+    @classmethod
+    def create(
+            cls,
+            element_type: ElementType,
+            display_name: str,
+            sort_criteria_builder: Callable[[bool], list[AlbumSortCriteria]],
+            descending: bool = False):
+        obj : FavoriteAlbumsMode = FavoriteAlbumsMode()
+        obj.__element_type = element_type
+        obj.__display_name = display_name
+        obj.__sort_criteria_builder = sort_criteria_builder
+        obj.__descending = descending
+        return obj
+
+    @property
+    def element_type(self) -> ElementType:
+        return self.__element_type
+
+    @property
+    def display_name(self) -> str:
+        return self.__display_name
+
+    @property
+    def sort_criteria_builder(self) -> Callable[[bool], list[AlbumSortCriteria]]:
+        return self.__sort_criteria_builder
+
+    @property
+    def descending(self) -> bool:
+        return self.__descending
+
+
 def get_image_dimension_list(obj : any) -> list[int]:
     key = type(obj).__name__
     return default_image_sz_by_type[key] if key in default_image_sz_by_type else list()
@@ -63,7 +108,68 @@ def get_name_or_title(obj : any) -> str:
     return None
 
 
-def get_image_url(obj : any) -> str:
+def set_album_art_from_album_id(album_id: str, tidal_session: TidalSession, entry: dict):
+    album_art_url: str = get_album_art_url_by_id(album_id=album_id, tidal_session=tidal_session)
+    upnp_util.set_album_art_from_uri(album_art_url, entry)
+
+
+def get_album_art_url_by_id(album_id: str, tidal_session: TidalSession) -> str:
+    if config.enable_image_caching:
+        # try cached!
+        document_root_dir : str = upmplgutils.getOptionValue("webserverdocumentroot")
+        if document_root_dir:
+            sub_dir_list: list[str] = [constants.plugin_name, "images", TidalAlbum.__name__]
+            image_dir: str = ensure_directory(document_root_dir, sub_dir_list)
+            cached_file_name: str = f"{str(album_id)}.jpg"
+            cached_file: str = os.path.join(image_dir, cached_file_name)
+            if os.path.exists(cached_file):
+                # use cached file
+                path: list[str] = list()
+                path.extend(sub_dir_list)
+                path.append(cached_file_name)
+                cached_image_url: str = compose_docroot_url("/".join(path))
+                if config.dump_image_caching:
+                    msgproc.log(f"get_album_art_url_by_id [{album_id}] -> [{cached_image_url}]")
+                return cached_image_url
+            else:
+                if config.dump_image_caching:
+                    msgproc.log(f"get_album_art_url_by_id [{album_id}] -> cache miss")
+    # if we are are, we fallback to normal
+    if config.dump_image_caching:
+        msgproc.log(f"get_album_art_url_by_id [{album_id}] -> loading from upstream service is required")
+    album: TidalAlbum = try_get_album(album_id=album_id, tidal_session=tidal_session)
+    return get_image_url(obj=album, refresh=True)
+
+
+def get_image_url(obj : any, refresh: bool = False) -> str:
+    if isinstance(obj, TidalTrack):
+        # use album instead
+        track: TidalTrack = obj
+        obj = track.album
+    if not config.enable_image_caching: return __get_image_url(obj)
+    document_root_dir : str = upmplgutils.getOptionValue("webserverdocumentroot")
+    # webserverdocumentroot is required
+    if not document_root_dir: return __get_image_url(obj)
+    if type(obj) not in [TidalAlbum, TidalArtist]:
+        return __get_image_url(obj)
+    sub_dir_list : list[str] = [constants.plugin_name, "images", type(obj).__name__]
+    image_dir : str = ensure_directory(document_root_dir, sub_dir_list)
+    cached_file_name: str = f"{str(obj.id)}.jpg"
+    cached_file: str = os.path.join(image_dir, cached_file_name)
+    if refresh or not os.path.exists(cached_file):
+        image_url : str = __get_image_url(obj=obj)
+        # msgproc.log(f"get_image_url saving to [{image_dir}] [{dest_file}]")
+        img_data : bytes = requests.get(image_url).content
+        with open(cached_file, 'wb') as handler:
+            handler.write(img_data)
+    path : list[str] = list()
+    path.extend(sub_dir_list)
+    path.append(cached_file_name)
+    cached_image_url: str = compose_docroot_url("/".join(path))
+    return cached_image_url
+
+
+def __get_image_url(obj : any) -> str:
     dimension_list : list[int] = get_image_dimension_list(obj)
     if not dimension_list or len(dimension_list) == 0:
         msgproc.log(f"Type [{type(obj).__name__}] does not have an image sizes list!")
@@ -100,6 +206,15 @@ def is_multidisc_album(album : TidalAlbum) -> bool:
     return album.num_volumes and album.num_volumes > 1
 
 
+def try_get_track(tidal_session : TidalSession, track_id : str) -> TidalTrack:
+    track : TidalTrack = None
+    try:
+        track = tidal_session.track(track_id)
+    except Exception as ex:
+        msgproc.log(f"try_get_track failed for track_id [{track_id}] [{type(ex)}] [{ex}]")
+    return track
+
+
 def try_get_album(tidal_session : TidalSession, album_id : str) -> TidalAlbum:
     album : TidalAlbum = None
     try:
@@ -109,6 +224,15 @@ def try_get_album(tidal_session : TidalSession, album_id : str) -> TidalAlbum:
     return album
 
 
+def try_get_artist(tidal_session : TidalSession, artist_id : str) -> TidalArtist:
+    artist : TidalArtist = None
+    try:
+        artist = tidal_session.artist(artist_id)
+    except Exception as ex:
+        msgproc.log(f"try_get_artist failed for artist_id [{artist_id}] [{type(ex)}] [{ex}]")
+    return artist
+
+
 class CachedTidalQuality:
 
     def __init__(
@@ -262,16 +386,42 @@ def album_playlist_action(
     return album
 
 
-def is_stereo(album : TidalAlbum) -> bool:
+def is_tidal_album_stereo(album : TidalAlbum) -> bool:
     # missing -> assume STEREO
-    media_metadata_tags : list[str] = album.media_metadata_tags
+    media_metadata_tags: list[str] = album.media_metadata_tags
+    return is_stereo(media_metadata_tags)
+
+
+def __audio_modes_has_stereo(audio_modes: list[str]) -> bool:
+    if not audio_modes or len(audio_modes) == 0: return False
+    current: str
+    for current in audio_modes:
+        if TidalAudioMode.stereo == current: return True
+    return False
+
+
+def is_stereo(media_metadata_tags : list[str]) -> bool:
+    # nothing available -> assume STEREO
     if not media_metadata_tags or len(media_metadata_tags) == 0: return True
-    return (TidalMediaMetadataTags.sony_360 not in media_metadata_tags and
-            TidalMediaMetadataTags.dolby_atmos not in media_metadata_tags)
+    # return (TidalMediaMetadataTags.sony_360 not in media_metadata_tags and
+    #         TidalMediaMetadataTags.dolby_atmos not in media_metadata_tags)
+    return not __only_of(
+        media_metadata_tags=media_metadata_tags,
+        hit=[
+            TidalMediaMetadataTags.sony_360,
+            TidalMediaMetadataTags.dolby_atmos])
 
 
-def not_stereo_skipmessage(album : TidalAlbum) -> str:
-    return f"Skipping album with id [{album.id}] because [{album.media_metadata_tags}]"
+def __only_of(media_metadata_tags : list[str], hit: list[str]) -> bool:
+    if not media_metadata_tags or len(media_metadata_tags) == 0: return False
+    current: str
+    for current in media_metadata_tags:
+        if current not in hit: return False
+    return True
+
+
+def not_stereo_skipmessage(album_id: str, media_metadata_tags : list[str]) -> str:
+    return f"Skipping album with id [{album_id}] because [{media_metadata_tags}]"
 
 
 def __is_mqa(media_metadata_tags : list[str]) -> bool:
@@ -289,16 +439,47 @@ def __get_best_quality(media_metadata_tags : list[str]) -> str:
     return None
 
 
+def try_get_all_favorites(tidal_session: TidalSession) -> list[TidalAlbum]:
+    favorite_list: list[TidalAlbum] = list()
+    offset : int = 0
+    limit : int = 100
+    while True:
+        some : list[TidalAlbum] = None
+        try:
+            some : list[TidalAlbum] = tidal_session.user.favorites.albums(limit=limit, offset=offset)
+        except Exception as ex:
+            msg: str = f"Cannot get favorite albums from offset [{offset}] [{type(ex)}] [{ex}]"
+            raise Exception(msg)
+        some_len: int = len(some) if some else 0
+        if some_len > 0: favorite_list.extend(some)
+        if some_len < limit:
+            break
+        # another slice maybe
+        offset += limit
+    return favorite_list
+
+
 def get_quality_badge(
         album : TidalAlbum,
         cached_tidal_quality : CachedTidalQuality) -> str:
-    audio_modes : list[str] = album.audio_modes
+    return get_quality_badge_raw(
+        audio_modes=album.audio_modes,
+        media_metadata_tags=album.media_metadata_tags,
+        audio_quality=album.audio_quality,
+        cached_tidal_quality=cached_tidal_quality)
+
+
+def get_quality_badge_raw(
+        audio_modes : list[str],
+        media_metadata_tags: list[str],
+        audio_quality : str,
+        cached_tidal_quality : CachedTidalQuality) -> str:
+    # msgproc.log(f"get_quality_badge type(audio_modes) -> {type(audio_modes) if audio_modes else 'none'}")
     # TODO maybe map DOLBY_ATMOS to say Atmos
-    if audio_modes and audio_modes[0] != "STEREO": return audio_modes[0]
-    media_metadata_tags : list[str] = album.media_metadata_tags
-    tidal_quality : TidalQuality = __get_best_quality(media_metadata_tags)
+    if audio_modes and not __audio_modes_has_stereo(audio_modes): return ",".join(audio_modes)
+    tidal_quality: TidalQuality = __get_best_quality(media_metadata_tags)
     is_mqa : bool = __is_mqa(media_metadata_tags)
-    if not tidal_quality: tidal_quality = album.audio_quality
+    if not tidal_quality: tidal_quality = audio_quality
     stream_info_available : bool = (cached_tidal_quality and
                 cached_tidal_quality.bit_depth and
                 cached_tidal_quality.sample_rate)
@@ -336,10 +517,11 @@ def get_quality_badge(
         badge = "Low/320"
     elif TidalQuality.low_96k == tidal_quality:
         badge = "Low/96"
-    msgproc.log(f"get_quality_badge q:[{tidal_quality}] "
-                f"b:[{bit_depth}] "
-                f"s:[{sample_rate}] "
-                f"mqa:[{is_mqa}] -> [{badge}]")
+    if config.display_quality_badge:
+        msgproc.log(f"get_quality_badge q:[{tidal_quality}] "
+                    f"b:[{bit_depth}] "
+                    f"s:[{sample_rate}] "
+                    f"mqa:[{is_mqa}] -> [{badge}]")
     return badge
 
 
@@ -379,7 +561,7 @@ def load_unique_ids_from_mix_or_playlist(
         previous_page_last_found_id : str = None,
         item_filter : Callable[[any], any] = lambda x : track_only(x),
         initial_offset : int = 0,
-        max_slice_size : int = 100) -> tuple[list[str], int]:
+        max_slice_size : int = 100) -> tuple[list[str], int, bool, str]:
     # msgproc.log(f"load_unique_ids_from_mix_or_playlist mix_or_playlist_id [{tidal_obj_id}] "
     #             f"tidal_obj_type [{tidal_obj_type}] "
     #             f"initial_offset [{initial_offset}] max_slice_size [{max_slice_size}]")
@@ -430,3 +612,48 @@ def load_unique_ids_from_mix_or_playlist(
             finished = True
             break
     return id_list, last_offset, finished, last_found
+
+
+def ensure_directory(base_dir : str, sub_dir_list : list[str]) -> str:
+    curr_sub_dir : str
+    curr_dir : str = base_dir
+    for curr_sub_dir in sub_dir_list:
+        new_dir : str = os.path.join(curr_dir, curr_sub_dir)
+        # msgproc.log(f"checking dir [{new_dir}] ...")
+        if not os.path.exists(new_dir):
+            msgproc.log(f"creating dir [{new_dir}] ...")
+            os.mkdir(new_dir)
+        # else:
+        #     msgproc.log(f"dir [{new_dir}] already exists.")
+        curr_dir = new_dir
+    return curr_dir
+
+
+def compose_docroot_url(right : str) -> str:
+    host_port : str = os.environ['UPMPD_UPNPHOSTPORT']
+    doc_root : str = os.environ['UPMPD_UPNPDOCROOT']
+    if not host_port and not doc_root: return None
+    return f"http://{host_port}/{right}"
+
+
+def get_oauth2_credentials_file_name() -> str:
+    return os.path.join(upmplgutils.getcachedir(constants.plugin_name), constants.oauth2_credentials_file_name)
+
+
+def oauth2_credential_file_exists() -> bool:
+    return os.path.exists(get_oauth2_credentials_file_name())
+
+
+def get_pkce_credentials_file_name() -> str:
+    return os.path.join(upmplgutils.getcachedir(constants.plugin_name), constants.pkce_credentials_file_name)
+
+
+def pkce_credential_file_exists() -> bool:
+    return os.path.exists(get_pkce_credentials_file_name())
+
+
+def is_instance_of_any(obj: any, type_list: list[type]) -> bool:
+    t: type
+    for t in type_list if type_list else list():
+        if isinstance(obj, t): return True
+    return False
