@@ -266,9 +266,9 @@ def _tracknofordoc(doc):
 
 # Detect '[disc N]' or '(disc N)' or ' disc N' at the end of an album title
 # Groups         1   2       3                 4                      5
-_albtitdnumre = "(.*)(\[disc ([0-9]+)\]|\(disc ([0-9]+)\)|[ ]+disc[ ]+([0-9]+))[ ]*$"
+_albtitdnumre = r"(.*)(\[disc ([0-9]+)\]|\(disc ([0-9]+)\)|[ ]+disc[ ]+([0-9]+))[ ]*$"
 _albtitdnumexp = re.compile(_albtitdnumre, flags=re.IGNORECASE)
-_folderdnumre = "(cd|disc)[ ]*([0-9]+)(.*)"
+_folderdnumre = r"(cd|disc)[ ]*([0-9]+)(.*)"
 _folderdnumexp = re.compile(_folderdnumre, flags=re.IGNORECASE)
 
 # Create album record for track if needed (not already there).
@@ -415,8 +415,7 @@ def _setalbumartists(conn):
 
 
 # Add albums to the recoll index so they can be searched for
-def _albumstorecoll(conn):
-    rcldb = recoll.connect(confdir=uprclinit.getRclConfdir(), writable=True)
+def _albumstorecoll(conn, rcldb):
     c = conn.cursor()
     #                 0        1          2          3         4         5
     stmt = '''SELECT album_id, albfolder, albtitle, albarturi, albdate, artist.value
@@ -425,7 +424,7 @@ def _albumstorecoll(conn):
     c.execute(stmt, ())
     for r in c:
         udi = "albid" + str(r[0])
-        doc = recoll.Doc()
+        doc = rcldb.doc()
         doc["album"] = r[2]
         doc["title"] = r[2]
         doc["mtype"] = "inode/directory"
@@ -438,15 +437,14 @@ def _albumstorecoll(conn):
     
 
 # Add artists to the recoll index so they can be searched for
-def _artiststorecoll(conn):
-    rcldb = recoll.connect(confdir=uprclinit.getRclConfdir(), writable=True)
+def _artiststorecoll(conn, rcldb):
     c = conn.cursor()
     #                 0         1
     stmt = "SELECT artist_id, value    FROM artist"
     c.execute(stmt, ())
     for r in c:
         udi = "artid" + str(r[0])
-        doc = recoll.Doc()
+        doc = rcldb.doc()
         doc["title"] = r[1]
         doc["mtype"] = "inode/directory"
         doc["url"] = "file://artists/" + r[1] # Not used ever
@@ -663,7 +661,8 @@ def recolltosql(conn, rcldocs):
     _createmergedalbums(conn)
     conn.commit()
     end = timer()
-    _albumstorecoll(conn)
-    _artiststorecoll(conn)
+    rcldb = recoll.connect(confdir=uprclinit.getRclConfdir(), writable=True)
+    _albumstorecoll(conn, rcldb)
+    _artiststorecoll(conn, rcldb)
     uplog(f"recolltosql: processed {totcnt} docs in {end-start:.1f} Seconds")
  
